from __future__ import print_function, division
import os, gzip
import numpy as np
import pandas as pd
import itertools
from ypy import memo, fs


# complementary bases (code borrowed liberally from ldsc repo)
COMPLEMENT = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}
# bases
BASES = COMPLEMENT.keys()
# true iff strand ambiguous
STRAND_AMBIGUOUS = {''.join(x): x[0] == COMPLEMENT[x[1]]
        for x in itertools.product(BASES, BASES)
        if x[0] != x[1]}
# SNPS we want to keep (pairs of alleles)
VALID_SNPS = {x for x in map(lambda y: ''.join(y), itertools.product(BASES, BASES))
        if x[0] != x[1] and not STRAND_AMBIGUOUS[x]}
# T iff SNP 1 has the same alleles as SNP 2 (allowing for strand or ref allele flip)
# and neither is invalid.
MATCH_ALLELES = {
        x for x in map(lambda y: ''.join(y), itertools.product(VALID_SNPS, VALID_SNPS))
        # strand and ref match
        if ((x[0] == x[2]) and (x[1] == x[3])) or
        # ref match, strand flip
        ((x[0] == COMPLEMENT[x[2]]) and (x[1] == COMPLEMENT[x[3]])) or
        # ref flip, strand match
        ((x[0] == x[3]) and (x[1] == x[2])) or
        # ref flip, strand flip
        ((x[0] == COMPLEMENT[x[3]]) and (x[1] == COMPLEMENT[x[2]]))}
# T iff SNP 1 has the same alleles as SNP 2 w/ ref allele flip and neither is invalid
FLIP_ALLELES = {x for x in MATCH_ALLELES
        # ref flip, strand match
        if ((x[0] == x[3]) and (x[1] == x[2])) or
        # ref flip, strand flip
        ((x[0] == COMPLEMENT[x[3]]) and (x[1] == COMPLEMENT[x[2]]))}

# Checks if SNP columns are equal. If so, save time by using concat instead of merge.
# y can be either a single df or a list of dfs
# if x is a list, then y is unnecessary
def smart_merge(x, y=[], how='inner', fail_if_nonmatching=False, drop_from_y=[], key='SNP'):
    # make y into a list if it's just a single df
    if type(y) == pd.DataFrame:
        y = [y]
    # if x is a list, pass the rest onto y
    if type(x) != pd.DataFrame: # assume x is a list
        y = x[1:] + y
        x = x[0]

    # drop uninteresting columns from y and check that snps match
    matching = True
    for d in y:
        d.drop(drop_from_y, axis=1, inplace=True)
        if len(x) != len(d) or (x[key] != d[key]).any():
            matching = False

    x = x.reset_index(drop=True)
    if matching:
        return pd.concat([x]+[d.reset_index(drop=True).drop(key, axis=1) for d in y],
                axis=1)
    else:
        if fail_if_nonmatching:
            print('smart_merge sees that the dataframs dont have same snps in same order')
            exit()
        else:
            for d in y:
                x = pd.merge(x,
                        d.reset_index(drop=True), how=how, on=key)
        return x

# keeps only snps in ref, flips alleles in df to match those in ref, and sets value at
# non-matching snps to a missing_val (0 by default). 
# ref must contain: SNP, A1, A2
# df must contain: SNP, A1, A2, and colnames
def reconciled_to(ref, df, colnames, othercolnames=[], signed=True, missing_val=0, key='SNP'):
    result = smart_merge(
        ref,
        df[[key,'A1','A2']+list(colnames)+othercolnames].rename(
            columns={'A1':'A1_df','A2':'A2_df'}),
        how='left',
        key=key)
    print(len(result), 'snps after merging')
    if len(result) != len(ref):
        print('WARNING: merged data frame is not the same length as reference data frame')
        print('   check for duplicate snps in one of the two dataframes')

    # snps in ref but not in df
    missing = result.A1_df.isnull()
    print('of', len(result), 'snps in merge,', missing.sum(), 'were missing in df')
    result.loc[missing,colnames] = missing_val
    result.loc[missing,'A1_df'] = result.loc[missing,'A2_df'] = '-'

    # snps in both, but either invalid or not matching
    a1234 = (result.A1+result.A2+result.A1_df+result.A2_df).apply(lambda y: y.upper())
    match = ~missing & a1234.apply(lambda y: y in MATCH_ALLELES)
    n_mismatch = (~missing & ~match).sum()
    print('of', (~missing).sum(), 'remaining snps,', n_mismatch,
            'are a) present in ref and df, b) do not have matching sets of alleles '+\
                    'that are both valid,')
    result.loc[~missing & ~match,colnames] = missing_val

    if signed:
        flip = match & a1234.apply(lambda y: y in FLIP_ALLELES)
        n_flip = flip.sum()
        print('of the remaining', match.sum(), 'snps,', n_flip, 'snps need flipping',
            'and', (match & ~flip).sum(), 'snps matched and did not need flipping')
        result.loc[flip,colnames] *= -1

    return result.drop(['A1_df', 'A2_df'], axis=1)


# Wrapper class for the annotation files used by ldsc and sldp
class Annotation(object):
    def __init__(self, stem_chr, signed=True):
        self.stem_chr = stem_chr
        self.signed = signed

    def filestem(self, chrnum='', mkdir=False):
        fname = '{}{}'.format(self.stem_chr, chrnum)
        if mkdir:
            fs.makedir_for_file(fname)
        return fname
    def annot_filename(self, chrnum):
        return self.filestem(chrnum) + '.annot.gz'
    def sannot_filename(self, chrnum, mkdir=False):
        return self.filestem(chrnum, mkdir) + '.sannot.gz'
    def info_filename(self, chrnum):
        return self.filestem(chrnum) + '.info'
    def ldscores_filename(self, chrnum):
        return self.filestem(chrnum) + '.l2.ldscore.gz'
    def RV_filename(self, chrnum, full=False):
        return self.filestem(chrnum) + '.RV.gz'

    def info_df(self, chrs=range(1,23)):
        if type(chrs) == int:
            return pd.read_csv(self.info_filename(chrs), sep='\t', index_col=0)
        else:
            return reduce(lambda x,y:x+y,
                    [self.info_df(c) for c in chrs])
    @memo.memoized
    def annot_df(self, chrnum):
        return pd.read_csv(self.annot_filename(chrnum),
                compression='gzip', header=0, sep='\t')
    @memo.memoized
    def sannot_df(self, chrnum):
        return pd.read_csv(self.sannot_filename(chrnum),
                compression='gzip', header=0, sep='\t')
    @memo.memoized
    def RV_df(self, chrnum):
        return pd.read_csv(self.RV_filename(chrnum), sep='\t')

    def names(self, chrnum, RV=False):
        if RV:
            fname = self.RV_filename(chrnum)
        else:
            fname = self.sannot_filename(chrnum)
        temp = pd.read_csv(fname, nrows=1, delim_whitespace=True)
        return [x for x in temp.columns.values if not(x in ['SNP','CHR','CM','BP','A1','A2'])]

    def total_sqnorms(self, chrs):
        return self.info_df(chrs)['sqnorm'].values
    def total_sizes(self, chrs):
        return self.info_df(chrs)['supp'].values


# basic testing
if __name__ == '__main__':
    a = Annotation('/groups/price/yakir/data/simannot/1000G3.wim5unm/mock/')
    print(a.names(22))
    print(a.sannot_df(22))
    print(a.total_sqnorms(22))
    print(a.total_sizes(22))
    print(a.total_sqnorms([1,22]))
    print(a.total_sizes([1,22]))

    ref = pd.read_csv('/groups/price/yakir/data/simannot/1000G3.wim5unm/mock/22.sannot.gz',
            delimiter='\t').drop('ANNOT', axis=1).iloc[:20]

    ref.ix[0,'A1'] = a.sannot_df(22).ix[0,'A2']
    ref.ix[0,'A2'] = a.sannot_df(22).ix[0,'A1']
    ref.ix[1,'A1'] = 'A'; ref.ix[1,'A2'] = 'C'
    print(ref)

    print(reconciled_to(ref, a.sannot_df(22), ['ANNOT']))



from __future__ import print_function, division
import numpy as np
from pysnptools.util import IntRangeSet


class SnpSubset(object):
    def __init__(self, snps_ucscbed, ucscbed):
        # use bedtools to create an indicator vector for the snps membership in the subset
        indicator = snps_ucscbed.intersect(ucscbed, c=True)
        self.irs = IntRangeSet(np.flatnonzero(
            np.array([int(snp.name) for snp in indicator])))

    def num_snps(self):
        return len(self.irs)

class SnpPartition(object):
    def __init__(self, snps_ucscbed, breakpoints_ucscbed, mhc_ucscbed=None):
        # use bedtools to find the snp indices of the breakpoints
        self.last_snps = np.array([])
        c = snps_ucscbed.closest(
                breakpoints_ucscbed.sort(), iu=True, D='ref').saveas()
        closest_breakpoints = np.array([' '.join(i[3:-1]) for i in c] + [''])
        indices = closest_breakpoints[:-1] != closest_breakpoints[1:]
        self.last_snps = np.flatnonzero(indices)

        # store MHC info if supplied
        self.mhc_ucscbed = mhc_ucscbed
        if self.mhc_ucscbed is not None:
            self.mhc = SnpSubset(snps_ucscbed, self.mhc_ucscbed)

    def ranges(self):
        ranges = zip(np.concatenate([[0], self.last_snps]), self.last_snps+1)
        if self.mhc_ucscbed is None:
            return ranges
        else:
            return [r for r in ranges if (IntRangeSet(r) & self.mhc.irs).isempty]

    def ind_ranges_overlapping(self, irs):
        ranges = self.ranges()
        return [i for i, r in enumerate(ranges) if not (IntRangeSet(r) & irs).isempty]


if __name__ == '__main__':
    import dataset as prd
    from pybedtools import BedTool
    d = prd.Dataset('/groups/price/yakir/data/datasets/1000G3.wim5u/1000G3.wim5u.')
    ldbreakpoints = BedTool(
            '/groups/price/yakir/data/reference/pickrell_breakpoints.hg19.eur.bed')
    sp = SnpPartition(d.ucscbed(22), ldbreakpoints)
    print(sp.ranges())
    print(sp.ind_ranges_overlapping(IntRangeSet('1190:2000')))



from __future__ import print_function, division
import os
import numpy as np
import pandas as pd
from pysnptools.snpreader import Bed
from pybedtools import BedTool
from ypy import memo
import ypy.iter as pyit
import time


# Wrapper class for plink-formatted genotype data sets
class Dataset(object):
    def __init__(self, bfile_chr, assembly='hg19'):
        self.bfile_chr=bfile_chr
        self.assembly=assembly

    @property
    def path(self):
        return os.path.dirname(self.bfile_chr) + '/'

    @memo.memoized
    def data(self, chrnum):
        return Bed(self.bfile(chrnum))
    def stdX(self, chrnum, r):
        return self.stdX_it(chrnum, range(r[0],r[1]))
    def stdX_it(self, chrnum, it):
        genotypes = self.data(chrnum)[:, it].read()
        genotypes.standardize(); genotypes.standardize()
        return genotypes.val

    def bfile(self, chrnum):
        return self.bfile_chr + str(chrnum)
    def bimfile(self, chrnum):
        return self.bfile(chrnum)+'.bim'
    def ucscbedfile(self, chrnum):
        return self.bfile(chrnum)+'.ucscbed'
    def frq_file(self, chrnum):
        return self.bfile(chrnum) + '.frq'
    @memo.memoized
    def frq_df(self, chrnum):
        return pd.read_csv(self.frq_file(chrnum), delim_whitespace=True, header=0)

    @memo.memoized
    def bim_df(self, chrnum):
        return pd.read_csv(self.bfile(chrnum)+'.bim',
                names=['CHR','SNP','CM','BP','A1','A2'],
                sep='\t')

    def M(self, chrnum):
        return len(self.bim_df(chrnum))
    def totalM(self, chroms=None):
        if chroms is None:
            chroms = [c for c in range(1,23) if os.path.exists(self.bimfile(c))]
        return sum(self.M(c) for c in chroms)
    def N(self):
        chrom = min([c for c in range(1,23) if os.path.exists(self.bimfile(c))])
        return self.data(chrom).iid_count

    @memo.memoized
    def ucscbed(self, chrnum):
        if os.path.exists(self.ucscbedfile(chrnum)):
            bt = BedTool(self.ucscbedfile(chrnum))
            if len(bt) == self.M(chrnum):
                return BedTool(self.ucscbedfile(chrnum))
            else:
                print('NOTE: ucscbed file for', self.bfile(chrnum), 'is wrong length.')
        print('NOTE: ucscbed file for',self.bfile(chrnum),'not found/incomplete. creating...')
        ucscbed = []
        for row in self.bim_df(chrnum).iterrows():
            ucscbed.append('chr{}\t{}\t{}'.format(
                row[1]['CHR'],
                row[1]['BP'],
                int(row[1]['BP'])+1))
        print('DONE')
        return BedTool('\n'.join(ucscbed), from_string=True).saveas(
                self.ucscbedfile(chrnum))

    # return ldblock info, standardized X, snp metadata
    def block_data(self, ldblocks, c, meta=None, chunksize=15, genos=True, verbose=2):
        # restrict to ld blocks in this chr and process them in chunks
        chr_blocks = ldblocks[ldblocks.chr=='chr'+str(c)]
        # get metadata about snps
        snps = self.bim_df(c)

        t0 = time.time()
        for block_nums in pyit.grouper(chunksize, range(len(chr_blocks))):
            # get ld blocks in this chunk, and indices of the snps that start and end them
            chunk_blocks = chr_blocks.iloc[block_nums]
            blockstarts_ind = np.searchsorted(snps.BP.values, chunk_blocks.start.values)
            blockends_ind = np.searchsorted(snps.BP.values, chunk_blocks.end.values)
            if verbose >= 1:
                print('{} : chr {} snps {} - {}'.format(
                    time.time()-t0, c, blockstarts_ind[0], blockends_ind[-1]))

            # read in refpanel for this chunk, and find the relevant annotated snps
            if genos:
                Xchunk = self.stdX(c, (blockstarts_ind[0], blockends_ind[-1]))
                print('read in chunk')
            else:
                Xchunk = None

            if meta is not None:
                metachunk = meta.iloc[blockstarts_ind[0]:blockends_ind[-1]]

            # calibrate ld block starts and ends with respect to the start of this chunk
            blockends_ind -= blockstarts_ind[0]
            blockstarts_ind -= blockstarts_ind[0]
            for i, start_ind, end_ind in zip(
                    chunk_blocks.index, blockstarts_ind, blockends_ind):
                if verbose >= 2:
                    print(time.time()-t0, ': processing ld block',
                            i, ',', end_ind-start_ind, 'snps')
                if genos:
                    X = Xchunk[:, start_ind:end_ind]
                else:
                    X = None
                if meta is not None:
                    metablock = metachunk.iloc[start_ind:end_ind]
                else:
                    metablock = None
                yield (chr_blocks.loc[i], X, metablock,
                    metachunk.iloc[start_ind:end_ind].index)


# basic tests
if __name__ == '__main__':
    d = Dataset('/groups/price/yakir/data/datasets/1000G3.wim5unm/1000G3.wim5unm.')
    print(d.path)
    print(d.bfile(22))
    print(d.frq_file(22))
    print(d.frq_df(22).columns)
    print(d.bim_df(22).columns)
    print(d.M(22))
    print(d.totalM())
    print(d.totalM(chroms=[21,22]))
    print(d.N())
